"""File gRPC client."""
from typing import Any, Dict, Generator
from uuid import UUID

import grpc
from fred_api.fileman import service_fileman_grpc_pb2_grpc
from fred_api.fileman.service_fileman_grpc_pb2 import ReadRequest, StatRequest
from frgal import GrpcClient, GrpcDecoder


class FileClient(GrpcClient):
    """File gRPC client."""

    grpc_service = 'File'

    def __init__(self, netloc: str, credentials: grpc.ChannelCredentials = None):
        """Initialize File client.

        Arguments:
            netloc: Network location of a gRPC server.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.

        Raises:
            FileNotFoundError: If the file is not found.
            ConnectionAbortedError: Another error in connection occured.
        """
        decoder = GrpcDecoder()
        super().__init__(netloc, decoder=decoder, grpc_modules=[service_fileman_grpc_pb2_grpc], credentials=credentials)

    def stat(self, uuid: UUID) -> Dict[str, Any]:
        """Get file metadata.

        Arguments:
            uuid: A file UUID.
        """
        request = StatRequest()
        request.uuid.value = str(uuid)
        try:
            return self.call(self.grpc_service, 'stat', request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise FileNotFoundError('File {} not found.'.format(uuid))
            raise ConnectionAbortedError('Connection failed: {}'.format(error))

    def read(self, uuid: UUID, size: int = 0) -> Generator[bytes, None, None]:
        """Get file content.

        Arguments:
            uuid: A file UUID.
            size: A maximum size of a chunk.

        Raises:
            FileNotFoundError: If the file is not found.
            ConnectionAbortedError: Another error in connection occured.
        """
        request = ReadRequest()
        request.uuid.value = str(uuid)
        request.size = size
        try:
            return self.call_stream(self.grpc_service, 'read', request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise FileNotFoundError('File {} not found.'.format(uuid))
            raise ConnectionAbortedError('Connection failed: {}'.format(error))
