"""High-level API for files from fileman service."""
from datetime import datetime
from io import SEEK_SET, BufferedIOBase, BytesIO, UnsupportedOperation
from typing import Optional
from uuid import UUID

from .client import FileClient


class File(BufferedIOBase):
    """Provide complete Binary I/O for a file in a fileman service."""

    def __init__(self, uuid: UUID, client: FileClient):
        self.uuid = uuid
        self.client = client
        # Fetch metadata
        self._stat = self.client.stat(self.uuid)
        # Internal buffer with file content
        self._buff = None  # type: Optional[BytesIO]

    @property
    def create_datetime(self) -> datetime:
        """Return file's create datetime."""
        return self._stat['create_datetime']

    @property
    def size(self) -> int:
        """Return file's size."""
        return self._stat['size']

    @property
    def mimetype(self) -> str:
        """Return file's MIME type."""
        return self._stat['mimetype']

    @property
    def name(self) -> str:
        """Return file's name."""
        return self._stat['name']

    @property
    def _raw(self) -> BytesIO:
        """Return the internal buffer. Create it if needed."""
        if self._buff is None:
            self._buff = BytesIO(b''.join(self.client.read(self.uuid)))
        return self._buff

    # Binary I/O API - IOBase https://docs.python.org/3/library/io.html#io.IOBase
    def close(self) -> None:
        """Raise UnsupportedOperation, this method is not supported."""
        raise UnsupportedOperation

    def readable(self) -> bool:
        """Return `True`, this stream can be read from."""
        return True

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        """Change the stream position to the given byte offset."""
        return self._raw.seek(offset, whence)

    def seekable(self) -> bool:
        """Return `False`, this stream doesn't support random access."""
        return True

    def writelines(self, lines) -> None:
        """Raise UnsupportedOperation, this method is not supported."""
        raise UnsupportedOperation

    # Binary I/O API - BufferedIOBase https://docs.python.org/3/library/io.html#io.BufferedIOBase

    def read(self, size=-1) -> bytes:
        """Read and return up to size bytes.

        If the argument is omitted, None, or negative, data is read and returned until EOF is reached.
        """
        return self._raw.read(size)


class Storage:
    """Utility storage to retrieve files from single service."""

    def __init__(self, client: FileClient):
        self.client = client

    def open(self, uuid: UUID):
        """Return a file with corresponding UUID."""
        return File(uuid, self.client)
