import types
from datetime import datetime
from unittest import TestCase
from unittest.mock import call, sentinel
from uuid import UUID

from fred_api.fileman.service_fileman_grpc_pb2 import ReadReply, ReadRequest, StatReply, StatRequest
from frgal.utils import TestClientMixin
from grpc import StatusCode
from grpc._channel import _RPCState, _SingleThreadedRendezvous as _Rendezvous
from pytz import UTC

from filed.client import FileClient


class TestFileClient(TestClientMixin, FileClient):
    """Testing version of a FileClient."""


class FileClientTest(TestCase):
    def setUp(self):
        self.client = TestFileClient(sentinel.netloc)

    def test_stat(self):
        uuid = UUID(int=42)
        name = 'example.txt'
        create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)
        mimetype = 'text/plain'
        size = 1024
        reply = StatReply()
        reply.data.create_datetime.FromDatetime(create_datetime)
        reply.data.size = size
        reply.data.mimetype = mimetype
        reply.data.name = name
        self.client.mock.return_value = reply

        result = self.client.stat(uuid)

        self.assertEqual(result, {'create_datetime': create_datetime, 'size': size, 'mimetype': mimetype, 'name': name})
        request = StatRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/stat', timeout=None)])

    def test_stat_does_not_exist(self):
        uuid = UUID(int=42)
        error = _Rendezvous(_RPCState((), '', '', StatusCode.NOT_FOUND, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(FileNotFoundError, 'File {} not found.'.format(uuid)):
            self.client.stat(uuid)

        request = StatRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/stat', timeout=None)])

    def test_stat_failed(self):
        uuid = UUID(int=42)
        error = _Rendezvous(_RPCState((), '', '', StatusCode.UNKNOWN, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(ConnectionAbortedError, 'Connection failed'):
            self.client.stat(uuid)

        request = StatRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/stat', timeout=None)])

    def test_read(self):
        uuid = UUID(int=42)
        content = b'Rimmer is a smeg head!'
        reply = ReadReply()
        reply.data.data = content
        self.client.mock.return_value = [reply]

        result = self.client.read(uuid)

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(tuple(result), (content, ))
        request = ReadRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/read', timeout=None)])

    def test_read_chunked(self):
        uuid = UUID(int=42)
        content = b'Rimmer is a smeg head!'
        reply1 = ReadReply()
        reply1.data.data = content[:12]
        reply2 = ReadReply()
        reply2.data.data = content[12:]
        self.client.mock.return_value = [reply1, reply2]

        result = self.client.read(uuid, size=12)

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(tuple(result), (content[:12], content[12:]))

        # Finish and check request
        request = ReadRequest()
        request.uuid.value = str(uuid)
        request.size = 12
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/read', timeout=None)])

    def test_read_does_not_exist(self):
        uuid = UUID(int=42)
        error = _Rendezvous(_RPCState((), '', '', StatusCode.NOT_FOUND, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(FileNotFoundError, 'File {} not found.'.format(uuid)):
            self.client.read(uuid)

        # Finish and check request
        request = ReadRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/read', timeout=None)])

    def test_read_failed(self):
        uuid = UUID(int=42)
        error = _Rendezvous(_RPCState((), '', '', StatusCode.UNKNOWN, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(ConnectionAbortedError, 'Connection failed'):
            self.client.read(uuid)

        # Finish and check request
        request = ReadRequest()
        request.uuid.value = str(uuid)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/read', timeout=None)])
