from datetime import datetime
from io import BytesIO, UnsupportedOperation
from typing import Any
from unittest import TestCase
from unittest.mock import call, sentinel
from uuid import UUID

from fred_api.fileman.service_fileman_grpc_pb2 import ReadReply, ReadRequest, StatReply, StatRequest
from grpc import StatusCode
from grpc._channel import _RPCState, _SingleThreadedRendezvous as _Rendezvous
from pytz import UTC

from filed.file import File, Storage

from .test_client import TestFileClient

A_UUID = UUID(int=42)
CONTENT = b"""Arnold Rimmer, Technician, 2nd Class.
Captain's remarks:
There's a saying amongst the officers:
"If a job's worth doing, it's worth doing well. If it's not worth doing, give it to Rimmer."
He aches for responsibility but constantly fails the engineering exam.
Astoundingly zealous.
Possibly mad.
Probably has more teeth than brain cells.
Promotion prospects: comical.
"""
CREATE_DATETIME = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)
SIZE = len(CONTENT)
MIMETYPE = 'text/plain'
NAME = 'rimmer.dat'


class FileTest(TestCase):
    def setUp(self):
        self.client = TestFileClient(sentinel.netloc)
        stat_reply = StatReply()
        stat_reply.data.create_datetime.FromDatetime(CREATE_DATETIME)
        stat_reply.data.size = SIZE
        stat_reply.data.mimetype = MIMETYPE
        stat_reply.data.name = NAME
        read_reply = ReadReply()
        read_reply.data.data = CONTENT
        self.client.mock.side_effect = [stat_reply, (read_reply, )]

    def assertStatCalled(self):
        request = StatRequest()
        request.uuid.value = str(A_UUID)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/stat', timeout=None)])

    def assertStatReadCalled(self):
        stat_request = StatRequest()
        stat_request.uuid.value = str(A_UUID)
        read_request = ReadRequest()
        read_request.uuid.value = str(A_UUID)
        calls = [call(stat_request, method='/Fred.Fileman.Api.File/stat', timeout=None),
                 call(read_request, method='/Fred.Fileman.Api.File/read', timeout=None)]
        self.assertEqual(self.client.mock.mock_calls, calls)

    def test_init(self):
        File(A_UUID, self.client)

        self.assertStatCalled()

    def test_init_not_found(self):
        error = _Rendezvous(_RPCState((), '', '', StatusCode.NOT_FOUND, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(FileNotFoundError, 'File {} not found.'.format(A_UUID)):
            File(A_UUID, self.client)

        self.assertStatCalled()

    def test_init_failed(self):
        error = _Rendezvous(_RPCState((), '', '', StatusCode.UNKNOWN, ""), None, None, None)
        self.client.mock.side_effect = error

        with self.assertRaisesRegex(ConnectionAbortedError, 'Connection failed'):
            File(A_UUID, self.client)

        self.assertStatCalled()

    def _test_metadata(self, name: str, value: Any):
        file = File(A_UUID, self.client)

        self.assertEqual(getattr(file, name), value)

        self.assertStatCalled()

    def test_create_datetime(self):
        self._test_metadata('create_datetime', CREATE_DATETIME)

    def test_size(self):
        self._test_metadata('size', SIZE)

    def test_mimetype(self):
        self._test_metadata('mimetype', MIMETYPE)

    def test_name(self):
        self._test_metadata('name', NAME)

    # Binary I/O API - IOBase https://docs.python.org/3/library/io.html#io.IOBase
    def test_close(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.close)

    def test_closed(self):
        file = File(A_UUID, self.client)

        self.assertFalse(file.closed)

    def test_fileno(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.fileno)

    def test_flush(self):
        file = File(A_UUID, self.client)

        file.flush()

    def test_isatty(self):
        file = File(A_UUID, self.client)

        self.assertFalse(file.isatty())

    def test_readable(self):
        file = File(A_UUID, self.client)

        self.assertTrue(file.readable())

    def test_readline(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.readline(), b'Arnold Rimmer, Technician, 2nd Class.\n')

        self.assertStatReadCalled()

    def test_readline_size(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.readline(10), b'Arnold Rim')

        self.assertStatReadCalled()

    def test_readlines(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.readlines(), CONTENT.splitlines(keepends=True))

        self.assertStatReadCalled()

    def test_readlines_hint(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.readlines(40), CONTENT.splitlines(keepends=True)[:2])

        self.assertStatReadCalled()

    def test_seek(self):
        file = File(A_UUID, self.client)

        file.seek(10)

        self.assertEqual(file.tell(), 10)
        self.assertStatReadCalled()

    def test_seekable(self):
        file = File(A_UUID, self.client)

        self.assertTrue(file.seekable())

    def test_tell(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.tell(), 0)

        self.assertStatReadCalled()

    def test_truncate(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.truncate)

    def test_writable(self):
        file = File(A_UUID, self.client)

        self.assertFalse(file.writable())

    def test_writelines(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.writelines, [])

    # Binary I/O API - BufferedIOBase https://docs.python.org/3/library/io.html#io.BufferedIOBase
    def test_detach(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.detach)

    def test_read(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.read(), CONTENT)

        self.assertStatReadCalled()

    def test_read_size(self):
        file = File(A_UUID, self.client)

        self.assertEqual(file.read(size=10), b'Arnold Rim')

        self.assertStatReadCalled()

    def test_read1(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.read1)

    def test_readinto(self):
        file = File(A_UUID, self.client)

        array = bytearray(10)
        self.assertEqual(file.readinto(array), 10)

        self.assertEqual(array, CONTENT[:10])
        self.assertStatReadCalled()

    def test_readinto1(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.readinto1, bytearray(10))

    def test_write(self):
        file = File(A_UUID, self.client)

        self.assertRaises(UnsupportedOperation, file.write, BytesIO())

    def test_read_failed(self):
        # Return a valid reply for stat, but fail for read.
        reply = StatReply()
        error = _Rendezvous(_RPCState((), '', '', StatusCode.UNKNOWN, ""), None, None, None)
        self.client.mock.side_effect = [reply, error]

        file = File(A_UUID, self.client)

        with self.assertRaisesRegex(ConnectionAbortedError, 'Connection failed'):
            file.read()

        self.assertStatReadCalled()


class StorageTest(TestCase):
    def setUp(self):
        self.client = TestFileClient(sentinel.netloc)

    def test_open(self):
        storage = Storage(self.client)
        reply = StatReply()
        reply.data.name = NAME
        self.client.mock.return_value = reply

        file = storage.open(A_UUID)

        self.assertEqual(file.name, NAME)
        request = StatRequest()
        request.uuid.value = str(A_UUID)
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Fileman.Api.File/stat', timeout=None)])
